import numpy as np
import math
import datetime
import re


class GcodeWriter:
    def __init__(self, z_up=2, z_down=-1, feedrate=800):
        self.gcode = "G54\n"
        self.z_up = z_up
        self.z_down = z_down
        self.xyz = (0, 0, 0)
        self.feedrate = feedrate

    def add_curve(self, xy_curve, z_curve=None, feedrate=None, lift=True):
        if z_curve is None:
            z_curve = self.z_down
        if feedrate is None:
            feedrate = self.feedrate

        x, y, z = self.xyz
        self.gcode += "F{:.2f}\n".format(feedrate)
        for i in range(len(xy_curve)):
            x, y = xy_curve[i]
            self.gcode += "G1 X{:.2f} Y{:.2f}\n".format(x, y)
            if i == 0:
                # first move, act on Z
                self.gcode += "G1 Z{:.2f}\n".format(z_curve)
                z = z_curve
        if lift:
            z = self.z_up
            self.gcode += "G1 Z{:.2f}\n".format(z)
        self.xyz = (x, y, z)

    def goto(self, x, y, z=None, feedrate=None):
        if feedrate is None:
            feedrate = self.feedrate
        x_c, y_c, z_c = self.xyz
        if z is None:
            z = z_c
        self.gcode += "G1 X{:.2f} Y{:.2f} Z{:.2f} F{:.2f}\n".format(x, y, z, feedrate)

    def save(self, filename):
        with open(filename, "w") as f:
            f.write(self.gcode)


def update_if_match(pattern, line, var):
    m = re.search(pattern, line)
    if m is None:
        return var
    else:
        return float(m.group(1))


def open_gcode(filename):
    with open(filename) as f:
        gcode = ""
        for line in f.readlines():
            gcode += line.rstrip() + "\n"
    return gcode


def insert_gcode(gcode1, gcode2, t=10, verbose=True):
    # inserts gcode2 in gcode1 at given interval
    gcode3 = ""

    # pos in mm (!)
    x = math.nan
    y = math.nan
    z = math.nan

    # feedrate in mm/min (!)
    f = 500

    dist = 0
    time = 0
    time_last = 0
    count = 0
    for line in gcode1.split("\n"):
        gcode3 += line+"\n"
        x_new = update_if_match(r"X([-\d.]+)", line, x)
        y_new = update_if_match(r"Y([-\d.]+)", line, y)
        z_new = update_if_match(r"Z([-\d.]+)", line, z)
        f = update_if_match(r"F([-\d.]+)", line, f)

        dx, dy, dz = (x_new-x), (y_new-y), (z_new-z)
        x, y, z = x_new, y_new, z_new
        delta_sq = map(lambda a: 0.0 if math.isnan(a) else a*a, (dx, dy, dz))
        delta = math.sqrt(sum(delta_sq))
        dist += delta
        time += delta*60/f
        if time-time_last > t:
            gcode3 += "(INSERT)\n"
            gcode3 += gcode2
            gcode3 += "(RETURN)\n"
            # return to current position
            gcode3 += "F{:.2f}\n".format(f)
            if not (math.isnan(x) or math.isnan(y)):
                gcode3 += "G1 X{:.2f} Y{:.2f}\n".format(x, y)
            if not math.isnan(z):
                gcode3 += "G1 Z{:.2f}\n".format(z)
            gcode3 += "(CONTINUE)\n"
            time_last = time
            count += 1
    if verbose:
        print("Total time: {}.".format(datetime.timedelta(seconds=round(time))))
        print("{} occurences inserted.".format(count))

    return gcode3
